import json
import asyncio
from .utils_file import Utils
import os
from .async_bucket import AsyncBucket
from aiohttp.client_exceptions import ClientResponseError
import logging

class Bucket():
    """
    This class deals with the GCP bucket.
    """
    def __init__(self,bucket_name : str ) -> None:
        if isinstance(bucket_name,str):
            self.__bucket_name = bucket_name
        else:
            raise Exception("Bucket name must be a string")

        self.__async_bucket = AsyncBucket()
        self.__utils = Utils()


    def download_files(self,files_names : list, folder : str) -> dict:
        """
        This function receives the list of files_list which are the files contained in files folder (new jobs).
        It tries to download each one of them from bucket. 
        If it succeeds it means that we already have the job.
        This helps to avoid duplicates.
        """

        if not files_names:
            raise Exception("files_names list must not be empty")
        elif not(isinstance(files_names,list) and all(isinstance(item, str) for item in files_names)):
            raise Exception("files names must be a list of strings containing the names of the files to download.")
        
     
        files_names = [folder+"/"+name for name in files_names]
        response = asyncio.run(self.__async_bucket.download(self.__bucket_name, files_names))
        fails = 0
        downloaded = 0
        for file in response[0]:
            try:
                if isinstance(file,ClientResponseError):
                    fails += 1
                else:
                    downloaded += 1
            except Exception as e:
                print(e)
                
        return {"response":response,"downloaded":downloaded,"fails":fails}



    def __compare_jobkeys(self, gcs_bucket_folder : str, comparison_field : str, files_path : str, filename_field : str) -> tuple([int,int]):
        """
        This function tries to download each file contained inside files folder. 
        If it downloads it means that we already have this offer in the bucket and in this case we only update the key.
        """

        files_names = list(os.listdir(files_path))

        #Generate dictionary with files that are already in the bucket and share the same jobkey as the files in files folder
        rs = self.download_files(files_names,folder=gcs_bucket_folder)
        response = rs["response"]
        downloaded_jobs_dict = self.__utils.parse_response(response,filename_field)



        updated = 0
        duplicates = 0
        #Iterate over files in files folder and update the key if we already have it in the bucket
        for filename in files_names:
            new_job = self.__utils.load_json_file(files_path + "/{}".format(filename)) #Scraped job offer
            
            #If this code fails it means that we don't have the job in the bucket, nothing needs to be done, we just upload it.
            try:
                downloaded_job = downloaded_jobs_dict[new_job[filename_field]] #Job offer from bucket

                #Iterate over the scraped offer esco id's and check if they are already inside of the one in the bucket
                #If not add and replace the file in the bucket with the updated esco id's
                new_esco_id = False
                for esco_id in new_job[comparison_field]:
                    if esco_id not in downloaded_job[comparison_field]:
                        downloaded_job[comparison_field].append(esco_id) #Append current id to the existing job
                        new_esco_id = True
                        updated += 1
                if new_esco_id:
                    self.__utils.save_json_file(downloaded_job,filename,files_path+"/") #Save the job with the new id
                else:
                    duplicates += 1
                    self.__utils.delete_file(filename,files_path) #Delete the file if we already have the offer

            except Exception as e:
                pass
        


        return updated, duplicates

    def upload_files(self,files : list, files_names : list, gcs_bucket_folder : str) -> int:
        """
        This function uploads the files contained in files_path folder to GCS bucket.
        It returns the number of files uploaded.
        """

        if not files:
            raise Exception("files list must not be empty")
        elif not files_names:
            raise Exception("files names list must not be empty")
        elif not(isinstance(files_names,list) and all(isinstance(item, str) for item in files_names)):
            raise Exception("files names must be a list of strings containing the names of the files to download.")
        elif len(files) != len(files_names):
            raise Exception("files and files_names must have the same length")

        files = [json.dumps(file) if isinstance(file,dict) else file for file in files]

        response = asyncio.run(self.__async_bucket.upload(self.__bucket_name, gcs_bucket_folder, files_names,files ))

        counter = 0
        for d in response[0]:
            try:
                if d["kind"] == "storage#object":
                    counter += 1
            except Exception as e:
                pass

        return counter

    def insert_replacing(self,doc_list : list, filename_field : str, bucket_folder : str,comparison_field : str, files_path : str = "temp/files") -> None:
        """
        This function compare each document from doc_list with the docs that are already in the bucket and uploads the new ones / updates the old ones.
        """


        if not(isinstance(doc_list,list) and all(isinstance(item, dict) for item in doc_list)):
            raise Exception("doc_list must be a list of dictionaries.")

        #If the folder doesn't exist create it
        if not os.path.exists(files_path):
            os.makedirs(files_path)

        ids = [doc_list[i][filename_field] +".json" for i in range(len(doc_list))] #Generator with Id's
        paths = [files_path+"/" for _ in ids] #Generator with paths
        fields = [comparison_field for _ in ids] #Generator with fields to compare


        self.__utils.clean_directory(files_path) 

        list(map(self.__utils.save_files_to_folder,doc_list,ids,paths,fields)) #Save jsons into files

        logging.info("{} unique documents.".format(len(list(os.listdir(files_path)))))

        updated, duplicates = self.__compare_jobkeys(
            gcs_bucket_folder=bucket_folder,
            comparison_field=comparison_field,
            files_path=files_path,
            filename_field=filename_field
            ) #Check if we already have those offers on bucket and update/delete them if we do

        files_names = list(os.listdir(files_path))

        files = []
        for file in files_names:
            with open(files_path+"/"+file, mode="r") as f:
                files.append(f.read())
        
        if files:
            files_inserted = self.upload_files(
                files=files,
                files_names=files_names,
                gcs_bucket_folder=bucket_folder,
                ) #Save files in files folder into the bucket

            logging.info("{} new files inserted.".format(files_inserted - updated))
            logging.info("{} updated files (same keys with different esco codes).".format(updated))
            logging.info("{} identical files not inserted.".format(duplicates))
        else:
            logging.info("{} new files inserted.".format(0))
            logging.info("{} updated files (same keys with different esco codes).".format(0))
            logging.info("{} identical files not inserted.".format(duplicates)) 

        self.__utils.clean_directory(files_path)


import os, shutil
import json
from aiohttp.client_exceptions import ClientResponseError

class Utils():
    def clean_directory(self,folder : str) -> None:
        """ This function clean the given directory """
        for filename in os.listdir(folder):
            file_path = os.path.join(folder, filename)
            try:
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.unlink(file_path)
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)
            except Exception as e:
                print('Failed to delete %s. Reason: %s' % (file_path, e))

    def save_json_file(self,json_data:dict,filename:str,folder:str) -> None:
        """
        This function save a json as a file from files folder
        """
        path = folder + filename

        with open(path, 'w') as f:
            json.dump(json_data, f)


    def delete_file(self,filename:str,files_path) -> None:
        """
        This function deletes a json as a file from files folder
        """
        os.remove(files_path + '/{}'.format(filename))

    def load_json_file(self,filename:str) -> dict:
        """Load a json file given a path"""
        with open(filename) as f:
            return json.load(f)

    def save_files_to_folder(self,json_data:dict, filename : str, folder : str, comparison_field : str) -> None:
        """
        This function saves the documents into files folder.
        It first checks if the files is already in the folder and if it is instead of replacing it it adds the new esco id's.
        """
        #Check if the file is already inside the folder if not just save it otherwise check if esco id's are the same and merge them if not
        path = folder + filename

        files = os.listdir(folder)
        if not filename in files:
            self.save_json_file(json_data,filename,folder)
        else:
            doc = self.load_json_file(path)
            if json_data[comparison_field] != doc[comparison_field]:
                doc[comparison_field] = list(set(doc[comparison_field]+ json_data[comparison_field]))
                self.save_json_file(doc,filename,folder)

    def parse_response(self,response: list, filename_field : str) -> list:
        """
        Receives a list of responses from GCS and returns a list of dictionaries with the information of each file.
        """        
        downloaded_jobs_dict = {}
        for file in response[0]:
            try:
                if isinstance(file,ClientResponseError):
                    continue
                else:
                    tmp_f = json.loads(file)
                    downloaded_jobs_dict.update({tmp_f[filename_field]:tmp_f})
            except Exception as e:
                print(e)
        return downloaded_jobs_dict



import backoff
import aiohttp
from gcloud.aio.storage import Storage
import logging

logging.getLogger('backoff').setLevel(logging.ERROR)


class StorageWithBackoff(Storage):
    @backoff.on_exception(backoff.expo, aiohttp.ClientResponseError,
                          max_tries=3, jitter=backoff.full_jitter)
    async def copy(self, *args, **kwargs):
        return await super().copy(*args, **kwargs)

    @backoff.on_exception(backoff.expo, aiohttp.ClientResponseError,
                          max_tries=1, jitter=backoff.full_jitter)
    async def download(self, *args, **kwargs):
        return await super().download(*args, **kwargs)

    @backoff.on_exception(backoff.expo, aiohttp.ClientResponseError,
                          max_tries=3, jitter=backoff.full_jitter)
    async def upload(self, *args, **kwargs):
        return await super().upload(*args, **kwargs)


from distutils.core import setup

setup(
    name = 'bucket_save_upload',
    packages = ['bucket_save_upload'],
    version = 'v1.0.2',
    description = "Save and upload to GCS buckets in a parallel fashion.",
    author="Rennan Valadares",
    author_email="rennanvoa2@gmail.com",
    url = 'https://github.com/NeobrainData/bucket-save',
    download_url = 'https://github.com/NeobrainData/bucket-save-upload/archive/refs/tags/v1.0.2.tar.gz',
    keywords = ['GCS', 'Google Cloud Storage', 'bucket', 'upload'],
    classifiers = [],
)


import asyncio
from gcloud.aio.auth import Token
from gcloud.aio.storage import Storage 
from .storage_backoff import StorageWithBackoff 
from aiohttp.client_exceptions import ClientConnectorError,ClientResponseError
class AsyncBucket():
    def __init__(self) -> None:
        # Used a token from a service account for authentication
        self.scopes = ["https://www.googleapis.com/auth/devstorage.full_control"]

    async def download(self,bucket : str, obj_names : list) -> list:
        """
        This function downloads a list of json objects from a bucket in a parallel fashion.
        """
        sa_token = Token(scopes=self.scopes)
        response = []
        async with Storage(token=sa_token) as client:
            tasks = (client.download(bucket, file) for file in obj_names) # Used the gcloud library download method, with required args
            try:
                response.append(await asyncio.gather(*tasks,return_exceptions=True))
            except ClientConnectorError as e:
                raise("Exception: ClientConnectorError.Please check GCS bucket name and credentials.")  
            except Exception as e:
                print(e)

        await sa_token.close()
        return response

    async def upload(self,bucket : str,gcs_bucket_folder : str, file_names : list, jsons : list) -> list:
        """
        This function uploads a list of objects to a bucket in a parallel fashion.
        """
        file_names = [gcs_bucket_folder+"/"+i for i in file_names]
        sa_token = Token(scopes=self.scopes)
        response = []
        async with StorageWithBackoff(token=sa_token) as client:
            tasks = [client.upload(*arg) for arg in zip([bucket for _ in range(len(file_names))],file_names,jsons)] # Used the gcloud library upload method, with required args
            try:
                response.append(await asyncio.gather(*tasks,return_exceptions=True))
            except ClientResponseError as e:
                raise Exception("Exception: ClientResponseError.Maybe due to a connection error.")

            except ClientConnectorError as e:
                raise Exception("Exception: ClientConnectorError.Please check GCS bucket name and credentials.")

            except Exception as e:
                if e == ClientConnectorError:
                    raise("Exception: ClientConnectorError.Please check GCS bucket name and credentials.")    
                else:
                    print(e)

        await sa_token.close()
        return response

